# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import json
from pathlib import Path

import json5
import pytest

# Variables
raillabel_v2_schema_path_var = (
    Path(__file__).parent.parent / "raillabel" / "schemas" / "raillabel_v2_schema.json"
)
metaschema_path_var = (
    Path(__file__).parent / "test_raillabel" / "__test_assets__" / "metaschema.json"
)
openlabel_v1_schema_path_var = (
    Path(__file__).parent / "test_raillabel" / "__test_assets__" / "openlabel_v1_schema.json"
)
openlabel_v1_short_path_var = (
    Path(__file__).parent / "test_raillabel" / "__test_assets__" / "openlabel_v1_short.json"
)
openlabel_v1_vcd_incompatible_path_var = (
    Path(__file__).parent
    / "test_raillabel"
    / "__test_assets__"
    / "openlabel_v1_vcd_incompatible.json"
)


@pytest.fixture(scope="session", autouse=True)
def compile_uncommented_test_file():
    """Compiles the main test file from json5 to json."""

    uncommented_example_file_path = Path(str(openlabel_v1_short_path_var) + "5")

    with uncommented_example_file_path.open() as f:
        data = json5.load(f)

    with openlabel_v1_short_path_var.open("w") as f:
        json.dump(data, f, indent=4)


# Raillabel v2 schema
@pytest.fixture
def raillabel_v2_schema_path():
    return raillabel_v2_schema_path_var


@pytest.fixture
def raillabel_v2_schema_data(request):
    raillabel_v2_schema_data = request.config.cache.get("raillabel_v2_schema_data", None)

    if raillabel_v2_schema_data is None:
        with raillabel_v2_schema_path_var.open() as raillabel_v2_short_file:
            raillabel_v2_schema_data = json.load(raillabel_v2_short_file)
            request.config.cache.set("raillabel_v2_schema_data", raillabel_v2_schema_data)

    return raillabel_v2_schema_data


# JSON metaschema
@pytest.fixture
def metaschema_path():
    return metaschema_path_var


@pytest.fixture
def metaschema_data(request):
    metaschema_data = request.config.cache.get("metaschema_data", None)

    if metaschema_data is None:
        with metaschema_path_var.open() as metashort_file:
            metaschema_data = json.load(metashort_file)
            request.config.cache.set("metaschema_data", metaschema_data)

    return metaschema_data


# JSON openlabel_v1_schema
@pytest.fixture
def openlabel_v1_schema_path():
    return openlabel_v1_schema_path_var


@pytest.fixture
def openlabel_v1_schema_data(request):
    openlabel_v1_schema_data = request.config.cache.get("openlabel_v1_schema_data", None)

    if openlabel_v1_schema_data is None:
        with openlabel_v1_schema_path_var.open() as openlabel_v1_short_file:
            openlabel_v1_schema_data = json.load(openlabel_v1_short_file)
            request.config.cache.set("openlabel_v1_schema_data", openlabel_v1_schema_data)

    return openlabel_v1_schema_data


# OpenLabel v1 short data
@pytest.fixture
def openlabel_v1_short_path():
    return openlabel_v1_short_path_var


@pytest.fixture
def openlabel_v1_short_data(request):
    openlabel_v1_short_data = request.config.cache.get("openlabel_v1_short_data", None)

    if openlabel_v1_short_data is None:
        with openlabel_v1_short_path_var.open() as openlabel_v1_short_file:
            openlabel_v1_short_data = json5.load(openlabel_v1_short_file)
            request.config.cache.set("openlabel_v1_short_data", openlabel_v1_short_data)

    return openlabel_v1_short_data


# OpenLabel v1 short vcd incompatible
@pytest.fixture
def openlabel_v1_vcd_incompatible_path():
    return openlabel_v1_vcd_incompatible_path_var


@pytest.fixture
def openlabel_v1_vcd_incompatible_data(request):
    openlabel_v1_vcd_incompatible_data = request.config.cache.get(
        "openlabel_v1_vcd_incompatible_data", None
    )

    if openlabel_v1_vcd_incompatible_data is None:
        with openlabel_v1_vcd_incompatible_path_var.open() as openlabel_v1_vcd_incompatible_file:
            openlabel_v1_vcd_incompatible_data = json.load(openlabel_v1_vcd_incompatible_file)
            request.config.cache.set(
                "openlabel_v1_vcd_incompatible_data",
                openlabel_v1_vcd_incompatible_data,
            )

    return openlabel_v1_vcd_incompatible_data



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import decimal
import typing as t
import uuid
from dataclasses import dataclass, field

from .num import Num
from .object_data import ObjectData
from .sensor_reference import SensorReference


@dataclass
class Frame:
    """A container of dynamic, timewise, information.

    Parameters
    ----------
    uid: int
        Number of the frame withing the annotation file. Must be unique.
    timestamp: decimal.Decimal, optional
        Timestamp containing the Unix epoch time of the frame with up to nanosecond precision.
    sensors: dict of raillabel.format.SensorReference, optional
        References to the sensors with frame specific information like timestamp and uri.
        Default is {}.
    data: dict, optional
        Dictionary containing data directly connected to the frame and not to anny object.
        Dictionary keys are the ID-strings of the variable the data belongs to. Default is {}.
    object_data: dict of raillabel.format.ObjectData, optional
        Dictionary containing the annotations per object. Dictionary keys are the object uids.
        Default is {}.

    Read-Only Attributes
    --------------------
    annotations: dict
        Dictionary containing all annotations of this frame, regardless of object or annotation
        type. Dictionary keys are annotation UIDs.
    """

    uid: int
    timestamp: t.Optional[decimal.Decimal] = None
    sensors: t.Dict[str, SensorReference] = field(default_factory=dict)
    data: t.Dict[str, Num] = field(default_factory=dict)
    object_data: t.Dict[uuid.UUID, ObjectData] = field(default_factory=dict)

    @property
    def annotations(self) -> t.Dict[uuid.UUID, t.Any]:
        """Return dict containing all annotations of this frame.

        Dictionary keys are annotation UIDs.
        """
        annotations = {}
        for object in self.object_data.values():
            annotations.update(object.annotations)

        return annotations

    def asdict(self) -> dict:
        """Export self as a dict compatible with the OpenLABEL schema.

        Returns
        -------
        dict_repr: dict
            Dict representation of this class instance.

        Raises
        ------
        ValueError
            if an attribute can not be converted to the type required by the OpenLabel schema.
        """

        dict_repr = {}

        if self.timestamp is not None or self.sensors != {}:
            dict_repr["frame_properties"] = {}

        if self.timestamp is not None:
            dict_repr["frame_properties"]["timestamp"] = str(self.timestamp)

        if self.sensors != {}:
            dict_repr["frame_properties"]["streams"] = {
                str(k): v.asdict() for k, v in self.sensors.items()
            }

        if self.data != {}:
            dict_repr["frame_properties"]["frame_data"] = {
                "num": [v.asdict() for v in self.data.values()]
            }

        if self.object_data != {}:
            dict_repr["objects"] = {str(k): v.asdict() for k, v in self.object_data.items()}

        return dict_repr

    def __eq__(self, other) -> bool:
        """Handel equal comparisons."""

        if not hasattr(other, "__dict__"):
            return False

        if len(self.__dict__) != len(other.__dict__):
            return False

        for attr in self.__dict__:

            if type(getattr(self, attr)) == type(self):
                if getattr(self, attr).uid != getattr(other, attr).uid:
                    return False

            else:
                if getattr(self, attr) != getattr(other, attr):
                    return False

        return True



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import decimal
import json
import typing as t
import uuid
from pathlib import Path

from .. import exceptions, format
from ._loader_abc import LoaderABC


class LoaderRailLabelV2(LoaderABC):
    """Loader class for the OpenLabel v1 annotation format.

    Attributes
    ----------
    scene: raillabel.Scene
        Loaded raillabel.Scene with the data.
    warnings: list[str]
        List of warning strings, that have been found during the execution of load().
    """

    scene: format.Scene
    warnings: t.List[str]

    SCHEMA_PATH: Path = Path(__file__).parent.parent / "schemas" / "raillabel_v2_schema.json"

    _OPENLABEL_CLASS_MAPPING = {
        "bbox": format.Bbox,
        "cuboid": format.Cuboid,
        "num": format.Num,
        "poly2d": format.Poly2d,
        "poly3d": format.Poly3d,
        "vec": format.Seg3d,
    }

    @property
    def subschema_version(self):
        """Return subschema version."""
        with self.SCHEMA_PATH.open() as schema_file:
            subschema_version = json.load(schema_file)["version"]
        return subschema_version

    def load(self, data: dict, validate: bool = True) -> format.Scene:
        """Load the data into a raillabel.Scene and return it.

        Parameters
        ----------
        data: dict
            A dictionary loaded from a JSON-file.
        validate: bool
            If True, the annotation data is validated via the respective schema. This is highly
            recommended, as not validating the data may lead to Errors during loading or while handling
            the scene. However, validating may increase the loading time. Default is True.

        Returns
        -------
        scene: raillabel.Scene
            The loaded scene with the data.

        Raises
        ------
        raillabel.exceptions.SchemaError
            if validate is True and the data does not validate against the schema.
        """

        self.warnings = []

        if validate:
            self.validate(data)

        data = self._prepare_data(data)

        self.scene = format.Scene(
            metadata=format.Metadata.fromdict(
                data_dict=data["metadata"], subschema_version=self.subschema_version
            )
        )

        self._check_sensor_completeness(data["coordinate_systems"], data["streams"])

        for stream_id in data["streams"]:
            self.scene.sensors[stream_id] = format.Sensor.fromdict(
                uid=stream_id,
                cs_raw=data["coordinate_systems"][stream_id],
                stream_raw=data["streams"][stream_id],
            )

        for object_id in data["objects"]:
            self.scene.objects[object_id] = format.Object.fromdict(
                data["objects"][object_id], object_id
            )

        self._load_frames(data)

        return self.scene

    def supports(self, data: dict) -> bool:
        """Determine if the loader is suitable for the data (lightweight).

        Parameters
        ----------
        data: dict
            A dictionary loaded from a JSON-file.

        Returns
        -------
        bool:
            If True, the Loader class is suitable for the data.
        """

        if "subschema_version" in data["openlabel"]["metadata"]:
            return (
                "openlabel" in data
                and "metadata" in data["openlabel"]
                and "schema_version" in data["openlabel"]["metadata"]
                and data["openlabel"]["metadata"]["subschema_version"].startswith("2.")
            )

        else:
            return (
                "openlabel" in data
                and "metadata" in data["openlabel"]
                and "schema_version" in data["openlabel"]["metadata"]
            )

    # === Sub-functions for better readibility --- #

    def _load_frames(self, data: dict):

        # Iterates over the frames
        for uid, frame in data["frames"].items():

            # frame.uid and frame.timestamp
            self.scene.frames[int(uid)] = format.Frame(uid=int(uid))

            if "frame_properties" not in frame:
                frame["frame_properties"] = {}

            if "timestamp" in frame["frame_properties"]:
                self.scene.frames[int(uid)].timestamp = decimal.Decimal(
                    frame["frame_properties"]["timestamp"]
                )

            # frame.sensors
            if "streams" in frame["frame_properties"]:

                for sensor_uid, sensor in frame["frame_properties"]["streams"].items():

                    # Older version store the stream timestamp under stream_sync. This adressed here.
                    if "stream_sync" in sensor["stream_properties"]:
                        sensor["stream_properties"]["sync"] = sensor["stream_properties"][
                            "stream_sync"
                        ]

                        warning_message = f"Deprecated field 'stream_sync' in frame {uid}. Please update file with raillable.save()."
                        if warning_message not in self.warnings:
                            self.warnings.append(warning_message)

                    try:
                        self.scene.frames[int(uid)].sensors[sensor_uid] = format.SensorReference(
                            sensor=self.scene.sensors[sensor_uid],
                            timestamp=decimal.Decimal(
                                sensor["stream_properties"]["sync"]["timestamp"]
                            ),
                        )

                    except KeyError:
                        self.warnings.append(
                            f"{sensor_uid} does not exist as a stream, but is referenced in the "
                            + f"sync of frame {uid}."
                        )

                    else:
                        if "uri" in sensor:
                            self.scene.frames[int(uid)].sensors[sensor_uid].uri = sensor["uri"]

            # frame.data
            if "frame_data" in frame["frame_properties"]:

                # Iterates over the annotation types
                for ann_type in frame["frame_properties"]["frame_data"]:

                    # Raises a warnings, if the annotation type is not supported
                    if ann_type not in self._OPENLABEL_CLASS_MAPPING:
                        self.warnings.append(
                            f"Annotation type {ann_type} (frame {uid}, object {obj_uid}) is "
                            + "currently not supported."
                        )
                        continue

                    # Collects the converted annotations
                    annotations = format.AnnotationContainer()
                    for ann_raw in frame["frame_properties"]["frame_data"][ann_type]:

                        # Older version have the annotation UUID stored in the 'name' field. This
                        # needs to be corrected first.
                        if not "uid" in ann_raw:
                            try:
                                ann_raw["uid"] = str(uuid.UUID(ann_raw["name"]))
                            except ValueError:
                                ann_raw["uid"] = str(uuid.uuid4())
                            else:
                                ann_raw["name"] = "general"

                        # Raises a warning, if a duplicate annotation is detected
                        if ann_raw["uid"] in annotations:
                            self.warnings.append(
                                f"Annotation '{ann_raw['uid']}' is contained more than one time "
                                + f"in frame '{uid}'. A new UID is beeing assigned."
                            )
                            ann_raw["uid"] = str(uuid.uuid4())

                        # Converts the annotation
                        (annotations[ann_raw["uid"]], w,) = self._OPENLABEL_CLASS_MAPPING[
                            ann_type
                        ].fromdict(ann_raw, self.scene.sensors)
                        self.warnings.extend(w)

                    # Allocates the annotations to the frame
                    self.scene.frames[int(uid)].data = annotations

            # Iterates over the objects in the frame
            if "objects" in frame:
                for obj_uid, obj_ann in frame["objects"].items():

                    obj_ann = obj_ann["object_data"]

                    # frame.object_data
                    try:
                        self.scene.frames[int(uid)].object_data[obj_uid] = format.ObjectData(
                            object=self.scene.objects[obj_uid]
                        )

                    except KeyError:
                        self.warnings.append(
                            f"{obj_uid} does not exist as an object, but is referenced in the object"
                            + f" annotation of frame {uid}."
                        )
                        continue

                    # Since there are a lot of annotation types, that all require unique methods for
                    # parsing the data from the OpenLABEL format, the parsing is handed off to the
                    # individual data classes via the fromdict() method. The mapping of the OpenLABEL
                    # annotation types to the classes is performend via the openlable_class_mapping
                    # dict.

                    # Iterates over the annotation types
                    for ann_type in obj_ann:

                        # Raises a warnings, if the annotation type is not supported
                        if ann_type not in self._OPENLABEL_CLASS_MAPPING:
                            self.warnings.append(
                                f"Annotation type {ann_type} (frame {uid}, object {obj_uid}) is "
                                + "currently not supported."
                            )
                            continue

                        # Collects the converted annotations
                        for ann_raw in obj_ann[ann_type]:

                            ann_raw = self._correct_annotation_name(ann_raw, ann_type, obj_uid)

                            # Older versions store the URI attribute in the annotation attributes.
                            # This needs to be corrected if it is the case.
                            if "attributes" in ann_raw and "text" in ann_raw["attributes"]:
                                for i, attr in enumerate(ann_raw["attributes"]["text"]):
                                    if attr["name"] == "uri":
                                        self.scene.frames[int(uid)].sensors[
                                            ann_raw["coordinate_system"]
                                        ].uri = attr["val"]
                                        del ann_raw["attributes"]["text"][i]
                                        break

                            # Raises a warning, if a duplicate annotation is detected
                            if (
                                ann_raw["uid"]
                                in self.scene.frames[int(uid)].object_data[obj_uid].annotations
                            ):
                                self.warnings.append(
                                    f"Annotation '{ann_raw['uid']}' is contained more than one "
                                    + f"time in frame '{uid}'. A new UID is beeing assigned."
                                )
                                ann_raw["uid"] = str(uuid.uuid4())

                            # Converts the annotation
                            (
                                self.scene.frames[int(uid)]
                                .object_data[obj_uid]
                                .annotations[ann_raw["uid"]],
                                w,
                            ) = self._OPENLABEL_CLASS_MAPPING[ann_type].fromdict(
                                ann_raw,
                                self.scene.sensors,
                            )
                            self.warnings.extend(w)

    def _prepare_data(self, data: dict) -> dict:
        """Add optional fields to dict and simplify interaction.

        Parameters
        ----------
        data : dict
            JSON data.

        Returns
        -------
        dict
            Enhanced JSON data.
        """

        if "coordinate_systems" not in data["openlabel"]:
            data["openlabel"]["coordinate_systems"] = {}

        if "streams" not in data["openlabel"]:
            data["openlabel"]["streams"] = {}

        if "objects" not in data["openlabel"]:
            data["openlabel"]["objects"] = {}

        if "frames" not in data["openlabel"]:
            data["openlabel"]["frames"] = {}

        return data["openlabel"]

    def _check_sensor_completeness(self, cs_data: dict, stream_data: dict):
        """Check for corresponding cs and stream completeness.

        Parameters
        ----------
        cs_data : dict
            Coordinate system data in the RailLabel format.
        stream_data : dict
            Stream data in the RailLabel format.

        Raises
        ------
        raillabel.exceptions.MissingCoordinateSystemError
            if a stream has no corresponding coordinate system.
        raillabel.exceptions.MissingStreamError
            if a coordinate system has no corresponding stream.
        raillabel.exceptions.UnsupportedParentError
            if a coordinate system has no corresponding stream.
        """

        for stream_uid in stream_data:
            if stream_uid not in cs_data:
                raise exceptions.MissingCoordinateSystemError(
                    f"Stream {stream_uid} has no corresponding coordinate system."
                )

        for cs_uid in cs_data:
            if cs_uid == "base":
                continue

            if cs_data[cs_uid]["parent"] != "base":
                raise exceptions.UnsupportedParentError(
                    f"Only 'base' is permitted as a parent for coordinate system {cs_uid}, "
                    + f"not {cs_data[cs_uid]['parent']}."
                )

            if cs_uid not in stream_data:
                raise exceptions.MissingStreamError(
                    f"Coordinate sytem {cs_uid} has no corresponding stream."
                )

    def _correct_annotation_name(
        self, ann_raw: dict, ann_type: str, obj_uid: str
    ) -> t.Tuple[dict, t.List[str]]:

        if "uid" not in ann_raw:
            try:
                ann_raw["uid"] = str(uuid.UUID(ann_raw["name"]))
            except ValueError:
                ann_raw["uid"] = str(uuid.uuid4())

        ann_raw[
            "name"
        ] = f"{ann_raw['coordinate_system']}__{ann_type}__{self.scene.objects[obj_uid].type}"

        return ann_raw



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import decimal
import typing as t
from dataclasses import dataclass

from .sensor import Sensor


@dataclass
class SensorReference:  # TODO
    """A reference to a sensor in a specific frame.

    Parameters
    ----------
    sensor: raillabel.format.Sensor
        The sensor this SensorReference corresponds to.
    timestamp: decimal.Decimal
        Timestamp containing the Unix epoch time of the sensor in a specific frame with up to
        nanosecond precision.
    uri: str, optional
        URI to the file corresponding to the frame recording in the particular frame. Default is
        None.
    """

    sensor: Sensor
    timestamp: decimal.Decimal
    uri: t.Optional[str] = None

    def asdict(self) -> dict:
        """Export self as a dict compatible with the OpenLABEL schema.

        Returns
        -------
        dict_repr: dict
            Dict representation of this class instance.

        Raises
        ------
        ValueError
            if an attribute can not be converted to the type required by the OpenLabel schema.
        """

        dict_repr = {"stream_properties": {"sync": {"timestamp": str(self.timestamp)}}}

        if self.uri is not None:
            dict_repr["uri"] = self.uri

        return dict_repr



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent))

import raillabel


def test_valid_file(openlabel_v1_short_data):
    validation_result = raillabel.validate(openlabel_v1_short_data)
    assert validation_result[0]


def test_file_one_type_error(openlabel_v1_short_data):
    openlabel_v1_short_data["openlabel"]["streams"]["lidar"]["uri"] = 42

    validation_result = raillabel.validate(openlabel_v1_short_data)

    assert not validation_result[0] and len(validation_result[1]) == 1


def test_file_two_type_errors(openlabel_v1_short_data):
    openlabel_v1_short_data["openlabel"]["streams"]["lidar"]["uri"] = 42
    openlabel_v1_short_data["openlabel"]["coordinate_systems"]["base"]["type"] = "invalid_value"

    validation_result = raillabel.validate(openlabel_v1_short_data)

    assert not validation_result[0] and len(validation_result[1]) == 2


def test_valid_file_path(openlabel_v1_short_data, raillabel_v2_schema_path):
    assert raillabel_v2_schema_path.is_file()
    validation_result = raillabel.validate(openlabel_v1_short_data, str(raillabel_v2_schema_path))
    assert validation_result[0]


def test_invalid_file_path(openlabel_v1_short_data, raillabel_v2_schema_path):
    raillabel_v2_schema_path = str(raillabel_v2_schema_path) + "_invalid"

    with pytest.raises(FileNotFoundError):
        raillabel.validate(openlabel_v1_short_data, raillabel_v2_schema_path)


def test_invalid_schema_key(openlabel_v1_short_data):
    with pytest.raises(FileNotFoundError):
        raillabel.validate(openlabel_v1_short_data, "invalid_schema")


# Executes the test if the file is called
if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear"])



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
from abc import ABC, abstractmethod, abstractproperty
from dataclasses import dataclass, field

from .sensor import Sensor


@dataclass
class _Annotation(ABC):

    uid: str
    name: str
    attributes: t.Dict[str, t.Union[int, float, bool, str, list]] = field(default_factory=dict)
    sensor: Sensor = None

    @property
    @abstractproperty
    def _REQ_FIELDS(self) -> t.List[str]:
        raise NotImplementedError

    # === Public Methods =====================================================

    @abstractmethod
    def asdict(self) -> t.Dict:
        """Export self as a dict compatible with the OpenLABEL schema.

        Returns
        -------
        dict_repr: dict
            Dict representation of this class instance.

        Raises
        ------
        ValueError
            if an attribute can not be converted to the type required by the OpenLabel schema.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def fromdict(self, data_dict: t.Dict, sensors: t.Dict) -> t.Tuple[t.Dict, list]:
        """Generate a Bbox object from a dictionary in the OpenLABEL format.

        Parameters
        ----------
        data_dict: dict
            OpenLABEL format dictionary containing the data for the annotation.
        sensors: dict
            Dictionary containing all sensors for the scene.

        Returns
        -------
        annotation: Bbox
            Converted annotation.
        warnings: list of str
            List of non-critical errors, that have occurred during the conversion.
        """
        raise NotImplementedError

    # === Private Methods ====================================================

    def _annotation_required_fields_asdict(self) -> t.Dict:
        """Return the required fields from the parent class to dict."""
        return {
            "uid": str(self.uid),
            "name": str(self.name),
        }

    def _annotation_optional_fields_asdict(self) -> t.Dict:
        """Return the optional fields from the parent class to dict."""

        dict_repr = {}

        if self.sensor is not None:
            dict_repr["coordinate_system"] = str(self.sensor.uid)

        if self.attributes != {}:
            dict_repr["attributes"] = {}

            for attr_name, attr_value in self.attributes.items():

                # Since the annotation stores the attributes in a collective
                # dictionary, they must be seperated by type in order to comply
                # with the OpenLabel format.

                if type(attr_value) == str:
                    attr_type = "text"

                elif type(attr_value) in [float, int]:
                    attr_type = "num"

                elif type(attr_value) == bool:
                    attr_type = "boolean"

                elif type(attr_value) in [list, tuple]:
                    attr_type = "vec"

                else:
                    raise TypeError(
                        f"Attribute type {attr_value.__class__.__name__} of {attr_value} is not "
                        + "supported. Supported types are str, float, int, bool, list, tuple."
                    )

                if attr_type not in dict_repr["attributes"]:
                    dict_repr["attributes"][attr_type] = []

                dict_repr["attributes"][attr_type].append({"name": attr_name, "val": attr_value})

        return dict_repr

    # === Special Methods ====================================================

    def __post_init__(self):
        """Check for required arguments after __init__.

        Inheritance in dataclasses has the flaw, that when the parent class has fields with
        defaults and the child class has fields without, a TypeError is raised due to required
        fields following optional ones. This is solved by making all fields in the child
        optional and checking for required fields in __post_init__().

        Raises
        ------
        TypeError
            If a required field has not been set.
        """

        for f in self._REQ_FIELDS:
            if getattr(self, f) is None:
                raise TypeError(f"{f} is a required argument for {self.__class__.__name__}")



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os

import jsonschema
import pytest


def test_metaschema_validation(raillabel_v2_schema_data, metaschema_data):
    assert jsonschema.validate(raillabel_v2_schema_data, metaschema_data) is None


def test_sample_data_validation_subschema(raillabel_v2_schema_data, openlabel_v1_short_data):
    assert jsonschema.validate(openlabel_v1_short_data, raillabel_v2_schema_data) is None


def test_sample_data_validation_superschema(openlabel_v1_schema_data, openlabel_v1_short_data):
    assert jsonschema.validate(openlabel_v1_short_data, openlabel_v1_schema_data) is None


# Executes the test if the file is called
if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear"])



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import decimal
import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent))

import raillabel


@pytest.fixture
def loader():
    return raillabel.format_loaders.LoaderRailLabelV2()


def test_supports_true(openlabel_v1_short_data, loader):
    assert loader.supports(openlabel_v1_short_data)


def test_supports_false(openlabel_v1_short_data, loader):
    openlabel_v1_short_data["openlabel"]["metadata"]["subschema_version"] = "3.0.0"
    assert not loader.supports(openlabel_v1_short_data)


def test_load_metadata(openlabel_v1_short_data, loader, raillabel_v2_schema_data):
    scene = loader.load(openlabel_v1_short_data, validate=False)

    assert scene.metadata.annotator == "test_annotator"
    assert scene.metadata.comment == "test_comment"
    assert scene.metadata.name == "test_project"
    assert scene.metadata.schema_version == "1.0.0"
    assert scene.metadata.tagged_file == "test_folder"
    assert scene.metadata.subschema_version == raillabel_v2_schema_data["version"]


def test_load_sensors(openlabel_v1_short_data, loader):
    scene = loader.load(openlabel_v1_short_data, validate=False)

    assert len(scene.sensors) == 3

    assert "rgb_middle" in scene.sensors
    assert scene.sensors["rgb_middle"].uid == "rgb_middle"
    assert scene.sensors["rgb_middle"].type == "camera"
    assert scene.sensors["rgb_middle"].uri == "/S1206063/image"
    assert scene.sensors["rgb_middle"].intrinsics.camera_matrix == (
        0.48,
        0,
        0.81,
        0,
        0,
        0.16,
        0.83,
        0,
        0,
        0,
        1,
        0,
    )
    assert scene.sensors["rgb_middle"].intrinsics.distortion == (
        0.49,
        0.69,
        0.31,
        0.81,
        0.99,
    )
    assert scene.sensors["rgb_middle"].intrinsics.width_px == 2464
    assert scene.sensors["rgb_middle"].intrinsics.height_px == 1600
    assert scene.sensors["rgb_middle"].extrinsics.pos == raillabel.format.Point3d(0, 1, 2)
    assert scene.sensors["rgb_middle"].extrinsics.quat == raillabel.format.Quaternion(
        0.97518507, -0.18529384, -0.05469746, -0.10811315
    )

    assert "ir_middle" in scene.sensors
    assert scene.sensors["ir_middle"].uid == "ir_middle"
    assert scene.sensors["ir_middle"].type == "camera"
    assert scene.sensors["ir_middle"].uri == "/A0001781/image"
    assert scene.sensors["ir_middle"].intrinsics.camera_matrix == (
        0.47,
        0,
        0.85,
        0,
        0,
        0.15,
        0.23,
        0,
        0,
        0,
        1,
        0,
    )
    assert scene.sensors["ir_middle"].intrinsics.distortion == (
        0.19,
        0.66,
        0.31,
        0.21,
        0.99,
    )
    assert scene.sensors["ir_middle"].intrinsics.width_px == 640
    assert scene.sensors["ir_middle"].intrinsics.height_px == 480
    assert scene.sensors["ir_middle"].extrinsics.pos == raillabel.format.Point3d(0, 2, 1)
    assert scene.sensors["ir_middle"].extrinsics.quat == raillabel.format.Quaternion(
        -0.64984101, -0.72563166, 0.18784818, -0.12600959
    )

    assert "lidar" in scene.sensors
    assert scene.sensors["lidar"].uid == "lidar"
    assert scene.sensors["lidar"].type == "lidar"
    assert scene.sensors["lidar"].uri == "/lidar_merged"
    assert scene.sensors["lidar"].extrinsics.pos == raillabel.format.Point3d(0, 0, 0)
    assert scene.sensors["lidar"].extrinsics.quat == raillabel.format.Quaternion(0, 0, 0, 1)


def test_load_objects(openlabel_v1_short_data, loader):
    # Removes the object data pointers from the example file so that it needs to be generated from the data
    for object in openlabel_v1_short_data["openlabel"]["objects"].values():
        del object["frame_intervals"]
        del object["object_data_pointers"]

    scene = loader.load(openlabel_v1_short_data, validate=False)

    assert len(scene.objects) == 3

    assert "b40ba3ad-0327-46ff-9c28-2506cfd6d934" in scene.objects
    assert (
        scene.objects["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].uid
        == "b40ba3ad-0327-46ff-9c28-2506cfd6d934"
    )
    assert scene.objects["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].name == "person_0000"
    assert scene.objects["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].type == "person"

    assert "6fe55546-0dd7-4e40-b6b4-bb7ea3445772" in scene.objects
    assert (
        scene.objects["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"].uid
        == "6fe55546-0dd7-4e40-b6b4-bb7ea3445772"
    )
    assert scene.objects["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"].name == "person_0001"
    assert scene.objects["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"].type == "person"

    assert "22dedd49-6dcb-413b-87ef-00ccfb532e98" in scene.objects
    assert (
        scene.objects["22dedd49-6dcb-413b-87ef-00ccfb532e98"].uid
        == "22dedd49-6dcb-413b-87ef-00ccfb532e98"
    )
    assert scene.objects["22dedd49-6dcb-413b-87ef-00ccfb532e98"].name == "train_0000"
    assert scene.objects["22dedd49-6dcb-413b-87ef-00ccfb532e98"].type == "train"


def test_load_frame0_metadata(openlabel_v1_short_data, loader):
    scene = loader.load(openlabel_v1_short_data, validate=False)

    assert 0 in scene.frames
    assert scene.frames[0].timestamp == decimal.Decimal("1632321743.134149")

    assert len(scene.frames[0].sensors) == 3
    assert "rgb_middle" in scene.frames[0].sensors
    assert scene.frames[0].sensors["rgb_middle"].sensor == scene.sensors["rgb_middle"]
    assert scene.frames[0].sensors["rgb_middle"].timestamp == decimal.Decimal(
        "1632321743.100000072"
    )
    assert scene.frames[0].sensors["rgb_middle"].uri == "rgb_test0.png"
    assert "ir_middle" in scene.frames[0].sensors
    assert scene.frames[0].sensors["ir_middle"].sensor == scene.sensors["ir_middle"]
    assert scene.frames[0].sensors["ir_middle"].timestamp == decimal.Decimal("1632321743.106000004")
    assert scene.frames[0].sensors["ir_middle"].uri == "ir_test0.png"
    assert "lidar" in scene.frames[0].sensors
    assert scene.frames[0].sensors["lidar"].sensor == scene.sensors["lidar"]
    assert scene.frames[0].sensors["lidar"].timestamp == decimal.Decimal("1632321743.134149")
    assert scene.frames[0].sensors["lidar"].uri == "lidar_test0.pcd"

    assert len(scene.frames[0].data) == 2
    assert "test_frame_data0" in scene.frames[0].data
    assert scene.frames[0].data["test_frame_data0"].uid == "a06fe567-29c7-475b-92a4-fbca64e671a7"
    assert scene.frames[0].data["a06fe567-29c7-475b-92a4-fbca64e671a7"].name == "test_frame_data0"
    assert scene.frames[0].data["test_frame_data0"].val == 53.1
    assert scene.frames[0].data["test_frame_data0"].sensor == scene.sensors["rgb_middle"]
    assert "test_frame_data1" in scene.frames[0].data
    assert scene.frames[0].data["test_frame_data1"].uid == "4bb95df7-a051-48a9-b77e-72f27ca43f64"
    assert scene.frames[0].data["4bb95df7-a051-48a9-b77e-72f27ca43f64"].name == "test_frame_data1"
    assert scene.frames[0].data["test_frame_data1"].val == 10
    assert scene.frames[0].data["test_frame_data1"].sensor == scene.sensors["lidar"]

    assert len(scene.frames[0].object_data) == 2
    assert "b40ba3ad-0327-46ff-9c28-2506cfd6d934" in scene.frames[0].object_data
    assert (
        scene.frames[0].object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].object
        == scene.objects["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
    )
    assert "22dedd49-6dcb-413b-87ef-00ccfb532e98" in scene.frames[0].object_data
    assert (
        scene.frames[0].object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"].object
        == scene.objects["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
    )


def test_load_frame0_bboxs(openlabel_v1_short_data, loader):
    scene = loader.load(openlabel_v1_short_data, validate=False)

    assert len(scene.frames[0].object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].bboxs) == 2

    assert (
        "78f0ad89-2750-4a30-9d66-44c9da73a714"
        in scene.frames[0].object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].bboxs
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .bboxs["78f0ad89-2750-4a30-9d66-44c9da73a714"]
        == scene.frames[0].annotations["78f0ad89-2750-4a30-9d66-44c9da73a714"]
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .bboxs["78f0ad89-2750-4a30-9d66-44c9da73a714"]
        .uid
        == "78f0ad89-2750-4a30-9d66-44c9da73a714"
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .bboxs["78f0ad89-2750-4a30-9d66-44c9da73a714"]
        .name
        == "rgb_middle__bbox__person"
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .bboxs["78f0ad89-2750-4a30-9d66-44c9da73a714"]
        .pos.x
        == 0
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .bboxs["78f0ad89-2750-4a30-9d66-44c9da73a714"]
        .pos.y
        == 1
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .bboxs["78f0ad89-2750-4a30-9d66-44c9da73a714"]
        .size.x
        == 2
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .bboxs["78f0ad89-2750-4a30-9d66-44c9da73a714"]
        .size.y
        == 3
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .bboxs["78f0ad89-2750-4a30-9d66-44c9da73a714"]
        .sensor
        == scene.sensors["rgb_middle"]
    )
    assert scene.frames[0].object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].bboxs[
        "78f0ad89-2750-4a30-9d66-44c9da73a714"
    ].attributes == {
        "test_text_attr0": "test_text_attr0_val",
        "test_text_attr1": "test_text_attr1_val",
        "test_num_attr0": 0,
        "test_num_attr1": 1,
        "test_bool_attr0": True,
        "test_vec_attr0": ["0", "1"],
        "test_vec_attr1": [0, 1, 2],
        "test_vec_attr2": ["a", "b", "c"],
    }

    assert (
        "68b4e02c-40c8-4de0-89ad-bc00ed05a043"
        in scene.frames[0].object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].bboxs
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .bboxs["68b4e02c-40c8-4de0-89ad-bc00ed05a043"]
        == scene.frames[0].annotations["68b4e02c-40c8-4de0-89ad-bc00ed05a043"]
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .bboxs["68b4e02c-40c8-4de0-89ad-bc00ed05a043"]
        .uid
        == "68b4e02c-40c8-4de0-89ad-bc00ed05a043"
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .bboxs["68b4e02c-40c8-4de0-89ad-bc00ed05a043"]
        .name
        == "ir_middle__bbox__person"
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .bboxs["68b4e02c-40c8-4de0-89ad-bc00ed05a043"]
        .pos.x
        == 3
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .bboxs["68b4e02c-40c8-4de0-89ad-bc00ed05a043"]
        .pos.y
        == 2
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .bboxs["68b4e02c-40c8-4de0-89ad-bc00ed05a043"]
        .size.x
        == 1
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .bboxs["68b4e02c-40c8-4de0-89ad-bc00ed05a043"]
        .size.y
        == 0
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .bboxs["68b4e02c-40c8-4de0-89ad-bc00ed05a043"]
        .sensor
        is scene.sensors["ir_middle"]
    )
    assert scene.frames[0].object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].bboxs[
        "68b4e02c-40c8-4de0-89ad-bc00ed05a043"
    ].attributes == {"test_bool_attr0": False}


def test_load_frame0_poly2ds(openlabel_v1_short_data, loader):
    scene = loader.load(openlabel_v1_short_data, validate=False)

    assert len(scene.frames[0].object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].poly2ds) == 2

    assert (
        "bebfbae4-61a2-4758-993c-efa846b050a5"
        in scene.frames[0].object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].poly2ds
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .poly2ds["bebfbae4-61a2-4758-993c-efa846b050a5"]
        == scene.frames[0].annotations["bebfbae4-61a2-4758-993c-efa846b050a5"]
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .poly2ds["bebfbae4-61a2-4758-993c-efa846b050a5"]
        .uid
        == "bebfbae4-61a2-4758-993c-efa846b050a5"
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .poly2ds["bebfbae4-61a2-4758-993c-efa846b050a5"]
        .name
        == "rgb_middle__poly2d__person"
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .poly2ds["bebfbae4-61a2-4758-993c-efa846b050a5"]
        .closed
        == False
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .poly2ds["bebfbae4-61a2-4758-993c-efa846b050a5"]
        .mode
        == "MODE_POLY2D_ABSOLUTE"
    )
    assert (
        len(
            scene.frames[0]
            .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
            .poly2ds["bebfbae4-61a2-4758-993c-efa846b050a5"]
            .points
        )
        == 2
    )
    assert scene.frames[0].object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].poly2ds[
        "bebfbae4-61a2-4758-993c-efa846b050a5"
    ].points[0] == raillabel.format.Point2d(1, 2)
    assert scene.frames[0].object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].poly2ds[
        "bebfbae4-61a2-4758-993c-efa846b050a5"
    ].points[1] == raillabel.format.Point2d(3, 4)
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .poly2ds["bebfbae4-61a2-4758-993c-efa846b050a5"]
        .sensor
        == scene.sensors["rgb_middle"]
    )
    assert scene.frames[0].object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].poly2ds[
        "bebfbae4-61a2-4758-993c-efa846b050a5"
    ].attributes == {
        "test_text_attr0": "test_text_attr0_val",
        "test_text_attr1": "test_text_attr1_val",
        "test_num_attr0": 2,
        "test_num_attr1": 1,
        "test_bool_attr0": True,
        "test_vec_attr0": ["0", "1"],
        "test_vec_attr1": [0, 1, 2],
        "test_vec_attr2": ["a", "b", "c"],
    }

    assert (
        "3f63201c-fb33-4487-aff6-ae0aa5fa976c"
        in scene.frames[0].object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].poly2ds
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .poly2ds["3f63201c-fb33-4487-aff6-ae0aa5fa976c"]
        == scene.frames[0].annotations["3f63201c-fb33-4487-aff6-ae0aa5fa976c"]
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .poly2ds["3f63201c-fb33-4487-aff6-ae0aa5fa976c"]
        .uid
        == "3f63201c-fb33-4487-aff6-ae0aa5fa976c"
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .poly2ds["3f63201c-fb33-4487-aff6-ae0aa5fa976c"]
        .name
        == "ir_middle__poly2d__person"
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .poly2ds["3f63201c-fb33-4487-aff6-ae0aa5fa976c"]
        .closed
        == True
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .poly2ds["3f63201c-fb33-4487-aff6-ae0aa5fa976c"]
        .mode
        == "MODE_POLY2D_ABSOLUTE"
    )
    assert (
        len(
            scene.frames[0]
            .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
            .poly2ds["3f63201c-fb33-4487-aff6-ae0aa5fa976c"]
            .points
        )
        == 4
    )
    assert scene.frames[0].object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].poly2ds[
        "3f63201c-fb33-4487-aff6-ae0aa5fa976c"
    ].points[0] == raillabel.format.Point2d(7, 6)
    assert scene.frames[0].object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].poly2ds[
        "3f63201c-fb33-4487-aff6-ae0aa5fa976c"
    ].points[1] == raillabel.format.Point2d(5, 4)
    assert scene.frames[0].object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].poly2ds[
        "3f63201c-fb33-4487-aff6-ae0aa5fa976c"
    ].points[2] == raillabel.format.Point2d(3, 2)
    assert scene.frames[0].object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].poly2ds[
        "3f63201c-fb33-4487-aff6-ae0aa5fa976c"
    ].points[3] == raillabel.format.Point2d(1, 0)
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .poly2ds["3f63201c-fb33-4487-aff6-ae0aa5fa976c"]
        .sensor
        == scene.sensors["ir_middle"]
    )
    assert scene.frames[0].object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].poly2ds[
        "3f63201c-fb33-4487-aff6-ae0aa5fa976c"
    ].attributes == {"test_bool_attr0": False}


def test_load_frame0_cuboids(openlabel_v1_short_data, loader):
    scene = loader.load(openlabel_v1_short_data, validate=False)

    assert len(scene.frames[0].object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].cuboids) == 2

    assert (
        "dc2be700-8ee4-45c4-9256-920b5d55c917"
        in scene.frames[0].object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].cuboids
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .cuboids["dc2be700-8ee4-45c4-9256-920b5d55c917"]
        == scene.frames[0].annotations["dc2be700-8ee4-45c4-9256-920b5d55c917"]
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .cuboids["dc2be700-8ee4-45c4-9256-920b5d55c917"]
        .uid
        == "dc2be700-8ee4-45c4-9256-920b5d55c917"
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .cuboids["dc2be700-8ee4-45c4-9256-920b5d55c917"]
        .name
        == "lidar__cuboid__person"
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .cuboids["dc2be700-8ee4-45c4-9256-920b5d55c917"]
        .pos.x
        == 0.49
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .cuboids["dc2be700-8ee4-45c4-9256-920b5d55c917"]
        .pos.y
        == 0.04
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .cuboids["dc2be700-8ee4-45c4-9256-920b5d55c917"]
        .pos.z
        == 0.73
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .cuboids["dc2be700-8ee4-45c4-9256-920b5d55c917"]
        .quat.x
        == 0
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .cuboids["dc2be700-8ee4-45c4-9256-920b5d55c917"]
        .quat.y
        == 0
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .cuboids["dc2be700-8ee4-45c4-9256-920b5d55c917"]
        .quat.z
        == 0
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .cuboids["dc2be700-8ee4-45c4-9256-920b5d55c917"]
        .quat.w
        == 0
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .cuboids["dc2be700-8ee4-45c4-9256-920b5d55c917"]
        .size.x
        == 0.75
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .cuboids["dc2be700-8ee4-45c4-9256-920b5d55c917"]
        .size.y
        == 0.01
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .cuboids["dc2be700-8ee4-45c4-9256-920b5d55c917"]
        .size.z
        == 0.10
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .cuboids["dc2be700-8ee4-45c4-9256-920b5d55c917"]
        .sensor
        == scene.sensors["lidar"]
    )
    assert scene.frames[0].object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].cuboids[
        "dc2be700-8ee4-45c4-9256-920b5d55c917"
    ].attributes == {
        "test_text_attr0": "test_text_attr0_val",
        "test_text_attr1": "test_text_attr1_val",
        "test_num_attr0": 0,
        "test_num_attr1": 1,
        "test_bool_attr0": True,
        "test_vec_attr0": ["0", "1"],
        "test_vec_attr1": [0, 1, 2],
        "test_vec_attr2": ["a", "b", "c"],
    }

    assert (
        "450ceb81-9778-4e63-bf89-42f3ed9f6747"
        in scene.frames[0].object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].cuboids
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .cuboids["450ceb81-9778-4e63-bf89-42f3ed9f6747"]
        == scene.frames[0].annotations["450ceb81-9778-4e63-bf89-42f3ed9f6747"]
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .cuboids["450ceb81-9778-4e63-bf89-42f3ed9f6747"]
        .uid
        == "450ceb81-9778-4e63-bf89-42f3ed9f6747"
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .cuboids["450ceb81-9778-4e63-bf89-42f3ed9f6747"]
        .name
        == "lidar__cuboid__person"
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .cuboids["450ceb81-9778-4e63-bf89-42f3ed9f6747"]
        .pos.x
        == 0.0
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .cuboids["450ceb81-9778-4e63-bf89-42f3ed9f6747"]
        .pos.y
        == 1.0
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .cuboids["450ceb81-9778-4e63-bf89-42f3ed9f6747"]
        .pos.z
        == 2.0
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .cuboids["450ceb81-9778-4e63-bf89-42f3ed9f6747"]
        .quat.x
        == 3.0
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .cuboids["450ceb81-9778-4e63-bf89-42f3ed9f6747"]
        .quat.y
        == 4.0
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .cuboids["450ceb81-9778-4e63-bf89-42f3ed9f6747"]
        .quat.z
        == 5.0
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .cuboids["450ceb81-9778-4e63-bf89-42f3ed9f6747"]
        .quat.w
        == 6.0
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .cuboids["450ceb81-9778-4e63-bf89-42f3ed9f6747"]
        .size.x
        == 7.0
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .cuboids["450ceb81-9778-4e63-bf89-42f3ed9f6747"]
        .size.y
        == 8.0
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .cuboids["450ceb81-9778-4e63-bf89-42f3ed9f6747"]
        .size.z
        == 9.0
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .cuboids["450ceb81-9778-4e63-bf89-42f3ed9f6747"]
        .sensor
        == scene.sensors["lidar"]
    )
    assert scene.frames[0].object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].cuboids[
        "450ceb81-9778-4e63-bf89-42f3ed9f6747"
    ].attributes == {"test_bool_attr0": False}


def test_load_frame0_seg3ds(openlabel_v1_short_data, loader):
    scene = loader.load(openlabel_v1_short_data, validate=False)

    assert len(scene.frames[0].object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].seg3ds) == 2

    assert (
        "c1087f1d-7271-4dee-83ad-519a4e3b78a8"
        in scene.frames[0].object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].seg3ds
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .seg3ds["c1087f1d-7271-4dee-83ad-519a4e3b78a8"]
        == scene.frames[0].annotations["c1087f1d-7271-4dee-83ad-519a4e3b78a8"]
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .seg3ds["c1087f1d-7271-4dee-83ad-519a4e3b78a8"]
        .uid
        == "c1087f1d-7271-4dee-83ad-519a4e3b78a8"
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .seg3ds["c1087f1d-7271-4dee-83ad-519a4e3b78a8"]
        .name
        == "lidar__vec__person"
    )
    assert scene.frames[0].object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].seg3ds[
        "c1087f1d-7271-4dee-83ad-519a4e3b78a8"
    ].point_ids == list(range(10))
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .seg3ds["c1087f1d-7271-4dee-83ad-519a4e3b78a8"]
        .sensor
        == scene.sensors["lidar"]
    )
    assert scene.frames[0].object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].seg3ds[
        "c1087f1d-7271-4dee-83ad-519a4e3b78a8"
    ].attributes == {
        "test_text_attr0": "test_text_attr0_val",
        "test_text_attr1": "test_text_attr1_val",
        "test_num_attr0": 0,
        "test_num_attr1": 1,
        "test_bool_attr0": True,
        "test_vec_attr0": ["0", "1"],
        "test_vec_attr1": [0, 1, 2],
        "test_vec_attr2": ["a", "b", "c"],
    }

    assert (
        "50be7fe3-1f43-47ca-b65a-930e6cfacfeb"
        in scene.frames[0].object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].seg3ds
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .seg3ds["50be7fe3-1f43-47ca-b65a-930e6cfacfeb"]
        == scene.frames[0].annotations["50be7fe3-1f43-47ca-b65a-930e6cfacfeb"]
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .seg3ds["50be7fe3-1f43-47ca-b65a-930e6cfacfeb"]
        .uid
        == "50be7fe3-1f43-47ca-b65a-930e6cfacfeb"
    )
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .seg3ds["50be7fe3-1f43-47ca-b65a-930e6cfacfeb"]
        .name
        == "lidar__vec__person"
    )
    assert scene.frames[0].object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].seg3ds[
        "50be7fe3-1f43-47ca-b65a-930e6cfacfeb"
    ].point_ids == list(reversed(range(10)))
    assert (
        scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .seg3ds["50be7fe3-1f43-47ca-b65a-930e6cfacfeb"]
        .sensor
        == scene.sensors["lidar"]
    )
    assert scene.frames[0].object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"].seg3ds[
        "50be7fe3-1f43-47ca-b65a-930e6cfacfeb"
    ].attributes == {"test_bool_attr0": False}


def test_load_frame0_poly3ds(openlabel_v1_short_data, loader):
    scene = loader.load(openlabel_v1_short_data, validate=False)

    assert len(scene.frames[0].object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"].poly3ds) == 1

    assert (
        "14f58fb0-add7-4ed9-85b3-74615986d854"
        in scene.frames[0].object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"].poly3ds
    )
    assert (
        scene.frames[0]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .poly3ds["14f58fb0-add7-4ed9-85b3-74615986d854"]
        == scene.frames[0].annotations["14f58fb0-add7-4ed9-85b3-74615986d854"]
    )
    assert (
        scene.frames[0]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .poly3ds["14f58fb0-add7-4ed9-85b3-74615986d854"]
        .uid
        == "14f58fb0-add7-4ed9-85b3-74615986d854"
    )
    assert (
        scene.frames[0]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .poly3ds["14f58fb0-add7-4ed9-85b3-74615986d854"]
        .name
        == "lidar__poly3d__train"
    )
    assert (
        scene.frames[0]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .poly3ds["14f58fb0-add7-4ed9-85b3-74615986d854"]
        .closed
        == False
    )
    assert (
        len(
            scene.frames[0]
            .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
            .poly3ds["14f58fb0-add7-4ed9-85b3-74615986d854"]
            .points
        )
        == 3
    )
    assert scene.frames[0].object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"].poly3ds[
        "14f58fb0-add7-4ed9-85b3-74615986d854"
    ].points[0] == raillabel.format.Point3d(9, 8, 7)
    assert scene.frames[0].object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"].poly3ds[
        "14f58fb0-add7-4ed9-85b3-74615986d854"
    ].points[1] == raillabel.format.Point3d(6, 5, 4)
    assert scene.frames[0].object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"].poly3ds[
        "14f58fb0-add7-4ed9-85b3-74615986d854"
    ].points[2] == raillabel.format.Point3d(3, 2, 1)
    assert (
        scene.frames[0]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .poly3ds["14f58fb0-add7-4ed9-85b3-74615986d854"]
        .sensor
        == scene.sensors["lidar"]
    )


def test_load_frame1_metadata(openlabel_v1_short_data, loader):
    scene = loader.load(openlabel_v1_short_data, validate=False)

    assert 1 in scene.frames
    assert scene.frames[1].timestamp == decimal.Decimal("1632321743.233263")

    assert len(scene.frames[1].sensors) == 3
    assert "rgb_middle" in scene.frames[1].sensors
    assert scene.frames[1].sensors["rgb_middle"].sensor == scene.sensors["rgb_middle"]
    assert scene.frames[1].sensors["rgb_middle"].timestamp == decimal.Decimal("1632321743.2")
    assert scene.frames[1].sensors["rgb_middle"].uri == "rgb_test1.png"
    assert "ir_middle" in scene.frames[1].sensors
    assert scene.frames[1].sensors["ir_middle"].sensor == scene.sensors["ir_middle"]
    assert scene.frames[1].sensors["ir_middle"].timestamp == decimal.Decimal("1632321743.208000004")
    assert scene.frames[1].sensors["ir_middle"].uri == "ir_test1.png"
    assert "lidar" in scene.frames[1].sensors
    assert scene.frames[1].sensors["lidar"].sensor == scene.sensors["lidar"]
    assert scene.frames[1].sensors["lidar"].timestamp == decimal.Decimal("1632321743.233263")
    assert scene.frames[1].sensors["lidar"].uri == "lidar_test1.pcd"

    assert len(scene.frames[1].data) == 2
    assert "test_frame_data0" in scene.frames[1].data
    assert scene.frames[1].data["test_frame_data0"].uid == "558697df-61f5-41b0-b112-3d6fcbd7d6c9"
    assert scene.frames[1].data["558697df-61f5-41b0-b112-3d6fcbd7d6c9"].name == "test_frame_data0"
    assert scene.frames[1].data["test_frame_data0"].val == 53
    assert scene.frames[1].data["test_frame_data0"].sensor == scene.sensors["rgb_middle"]
    assert "test_frame_data1" in scene.frames[1].data
    assert scene.frames[1].data["test_frame_data1"].uid == "843e07a0-aac5-4f62-8200-1468dbd3055d"
    assert scene.frames[1].data["843e07a0-aac5-4f62-8200-1468dbd3055d"].name == "test_frame_data1"
    assert scene.frames[1].data["test_frame_data1"].val == 10.1
    assert scene.frames[1].data["test_frame_data1"].sensor == scene.sensors["lidar"]

    assert len(scene.frames[1].object_data) == 2
    assert "6fe55546-0dd7-4e40-b6b4-bb7ea3445772" in scene.frames[1].object_data
    assert (
        scene.frames[1].object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"].object
        == scene.objects["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"]
    )
    assert "22dedd49-6dcb-413b-87ef-00ccfb532e98" in scene.frames[1].object_data
    assert (
        scene.frames[1].object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"].object
        == scene.objects["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
    )


def test_load_frame1_bboxs(openlabel_v1_short_data, loader):
    scene = loader.load(openlabel_v1_short_data, validate=False)

    assert len(scene.frames[1].object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"].bboxs) == 2

    assert (
        "6ba42cbc-484e-4b8d-a022-b23c2bb6643c"
        in scene.frames[1].object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"].bboxs
    )
    assert (
        scene.frames[1]
        .object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"]
        .bboxs["6ba42cbc-484e-4b8d-a022-b23c2bb6643c"]
        == scene.frames[1].annotations["6ba42cbc-484e-4b8d-a022-b23c2bb6643c"]
    )
    assert (
        scene.frames[1]
        .object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"]
        .bboxs["6ba42cbc-484e-4b8d-a022-b23c2bb6643c"]
        .uid
        == "6ba42cbc-484e-4b8d-a022-b23c2bb6643c"
    )
    assert (
        scene.frames[1]
        .object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"]
        .bboxs["6ba42cbc-484e-4b8d-a022-b23c2bb6643c"]
        .name
        == "rgb_middle__bbox__person"
    )
    assert (
        scene.frames[1]
        .object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"]
        .bboxs["6ba42cbc-484e-4b8d-a022-b23c2bb6643c"]
        .pos.x
        == 0
    )
    assert (
        scene.frames[1]
        .object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"]
        .bboxs["6ba42cbc-484e-4b8d-a022-b23c2bb6643c"]
        .pos.y
        == 1
    )
    assert (
        scene.frames[1]
        .object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"]
        .bboxs["6ba42cbc-484e-4b8d-a022-b23c2bb6643c"]
        .size.x
        == 2
    )
    assert (
        scene.frames[1]
        .object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"]
        .bboxs["6ba42cbc-484e-4b8d-a022-b23c2bb6643c"]
        .size.y
        == 3
    )
    assert (
        scene.frames[1]
        .object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"]
        .bboxs["6ba42cbc-484e-4b8d-a022-b23c2bb6643c"]
        .sensor
        == scene.sensors["rgb_middle"]
    )
    assert scene.frames[1].object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"].bboxs[
        "6ba42cbc-484e-4b8d-a022-b23c2bb6643c"
    ].attributes == {
        "test_text_attr0": "test_text_attr0_val",
        "test_text_attr1": "test_text_attr1_val",
        "test_num_attr0": 0,
        "test_num_attr1": 1,
        "test_bool_attr0": True,
        "test_vec_attr0": ["0", "1"],
        "test_vec_attr1": [0, 1, 2],
        "test_vec_attr2": ["a", "b", "c"],
    }

    assert (
        "5f28fa18-8f2a-4a40-a0b6-c0bbedc00f2e"
        in scene.frames[1].object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"].bboxs
    )
    assert (
        scene.frames[1]
        .object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"]
        .bboxs["5f28fa18-8f2a-4a40-a0b6-c0bbedc00f2e"]
        == scene.frames[1].annotations["5f28fa18-8f2a-4a40-a0b6-c0bbedc00f2e"]
    )
    assert (
        scene.frames[1]
        .object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"]
        .bboxs["5f28fa18-8f2a-4a40-a0b6-c0bbedc00f2e"]
        .uid
        == "5f28fa18-8f2a-4a40-a0b6-c0bbedc00f2e"
    )
    assert (
        scene.frames[1]
        .object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"]
        .bboxs["5f28fa18-8f2a-4a40-a0b6-c0bbedc00f2e"]
        .name
        == "ir_middle__bbox__person"
    )
    assert (
        scene.frames[1]
        .object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"]
        .bboxs["5f28fa18-8f2a-4a40-a0b6-c0bbedc00f2e"]
        .pos.x
        == 3
    )
    assert (
        scene.frames[1]
        .object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"]
        .bboxs["5f28fa18-8f2a-4a40-a0b6-c0bbedc00f2e"]
        .pos.y
        == 2
    )
    assert (
        scene.frames[1]
        .object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"]
        .bboxs["5f28fa18-8f2a-4a40-a0b6-c0bbedc00f2e"]
        .size.x
        == 1
    )
    assert (
        scene.frames[1]
        .object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"]
        .bboxs["5f28fa18-8f2a-4a40-a0b6-c0bbedc00f2e"]
        .size.y
        == 0
    )
    assert (
        scene.frames[1]
        .object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"]
        .bboxs["5f28fa18-8f2a-4a40-a0b6-c0bbedc00f2e"]
        .sensor
        == scene.sensors["ir_middle"]
    )
    assert scene.frames[1].object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"].bboxs[
        "5f28fa18-8f2a-4a40-a0b6-c0bbedc00f2e"
    ].attributes == {"test_bool_attr0": False}


def test_load_frame1_poly2ds(openlabel_v1_short_data, loader):
    scene = loader.load(openlabel_v1_short_data, validate=False)

    assert len(scene.frames[1].object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"].poly2ds) == 1

    assert (
        "e2503c5d-9fe4-4666-b510-ef644c5a766b"
        in scene.frames[1].object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"].poly2ds
    )
    assert (
        scene.frames[1]
        .object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"]
        .poly2ds["e2503c5d-9fe4-4666-b510-ef644c5a766b"]
        == scene.frames[1].annotations["e2503c5d-9fe4-4666-b510-ef644c5a766b"]
    )
    assert (
        scene.frames[1]
        .object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"]
        .poly2ds["e2503c5d-9fe4-4666-b510-ef644c5a766b"]
        .uid
        == "e2503c5d-9fe4-4666-b510-ef644c5a766b"
    )
    assert (
        scene.frames[1]
        .object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"]
        .poly2ds["e2503c5d-9fe4-4666-b510-ef644c5a766b"]
        .name
        == "rgb_middle__poly2d__person"
    )
    assert (
        scene.frames[1]
        .object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"]
        .poly2ds["e2503c5d-9fe4-4666-b510-ef644c5a766b"]
        .closed
        == False
    )
    assert (
        scene.frames[1]
        .object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"]
        .poly2ds["e2503c5d-9fe4-4666-b510-ef644c5a766b"]
        .mode
        == "MODE_POLY2D_ABSOLUTE"
    )
    assert (
        len(
            scene.frames[1]
            .object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"]
            .poly2ds["e2503c5d-9fe4-4666-b510-ef644c5a766b"]
            .points
        )
        == 2
    )
    assert scene.frames[1].object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"].poly2ds[
        "e2503c5d-9fe4-4666-b510-ef644c5a766b"
    ].points[0] == raillabel.format.Point2d(1, 2)
    assert scene.frames[1].object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"].poly2ds[
        "e2503c5d-9fe4-4666-b510-ef644c5a766b"
    ].points[1] == raillabel.format.Point2d(3, 4)
    assert (
        scene.frames[1]
        .object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"]
        .poly2ds["e2503c5d-9fe4-4666-b510-ef644c5a766b"]
        .sensor
        == scene.sensors["rgb_middle"]
    )
    assert scene.frames[1].object_data["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"].poly2ds[
        "e2503c5d-9fe4-4666-b510-ef644c5a766b"
    ].attributes == {
        "test_text_attr0": "test_text_attr0_val",
        "test_text_attr1": "test_text_attr1_val",
        "test_num_attr0": 0,
        "test_num_attr1": 1,
        "test_bool_attr0": True,
        "test_vec_attr0": ["0", "1"],
        "test_vec_attr1": [0, 1, 2],
        "test_vec_attr2": ["a", "b", "c"],
    }


def test_load_frame1_cuboids(openlabel_v1_short_data, loader):
    scene = loader.load(openlabel_v1_short_data, validate=False)

    assert len(scene.frames[1].object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"].cuboids) == 2

    assert (
        "536ac83a-32c8-4fce-8499-ef32716c64a6"
        in scene.frames[1].object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"].cuboids
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .cuboids["536ac83a-32c8-4fce-8499-ef32716c64a6"]
        == scene.frames[1].annotations["536ac83a-32c8-4fce-8499-ef32716c64a6"]
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .cuboids["536ac83a-32c8-4fce-8499-ef32716c64a6"]
        .uid
        == "536ac83a-32c8-4fce-8499-ef32716c64a6"
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .cuboids["536ac83a-32c8-4fce-8499-ef32716c64a6"]
        .name
        == "lidar__cuboid__train"
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .cuboids["536ac83a-32c8-4fce-8499-ef32716c64a6"]
        .pos.x
        == 0
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .cuboids["536ac83a-32c8-4fce-8499-ef32716c64a6"]
        .pos.y
        == 1
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .cuboids["536ac83a-32c8-4fce-8499-ef32716c64a6"]
        .pos.z
        == 2
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .cuboids["536ac83a-32c8-4fce-8499-ef32716c64a6"]
        .quat.x
        == 3
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .cuboids["536ac83a-32c8-4fce-8499-ef32716c64a6"]
        .quat.y
        == 4
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .cuboids["536ac83a-32c8-4fce-8499-ef32716c64a6"]
        .quat.z
        == 5
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .cuboids["536ac83a-32c8-4fce-8499-ef32716c64a6"]
        .quat.w
        == 6
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .cuboids["536ac83a-32c8-4fce-8499-ef32716c64a6"]
        .size.x
        == 7
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .cuboids["536ac83a-32c8-4fce-8499-ef32716c64a6"]
        .size.y
        == 8
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .cuboids["536ac83a-32c8-4fce-8499-ef32716c64a6"]
        .size.z
        == 9
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .cuboids["536ac83a-32c8-4fce-8499-ef32716c64a6"]
        .sensor
        == scene.sensors["lidar"]
    )
    assert scene.frames[1].object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"].cuboids[
        "536ac83a-32c8-4fce-8499-ef32716c64a6"
    ].attributes == {
        "test_text_attr0": "test_text_attr0_val",
        "test_text_attr1": "test_text_attr1_val",
        "test_num_attr0": 0,
        "test_num_attr1": 1,
        "test_bool_attr0": True,
        "test_vec_attr0": ["0", "1"],
        "test_vec_attr1": [0, 1, 2],
        "test_vec_attr2": ["a", "b", "c"],
    }

    assert (
        "e53bd5e3-980a-4fa7-a0f9-5a2e59ba663c"
        in scene.frames[1].object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"].cuboids
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .cuboids["e53bd5e3-980a-4fa7-a0f9-5a2e59ba663c"]
        == scene.frames[1].annotations["e53bd5e3-980a-4fa7-a0f9-5a2e59ba663c"]
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .cuboids["e53bd5e3-980a-4fa7-a0f9-5a2e59ba663c"]
        .uid
        == "e53bd5e3-980a-4fa7-a0f9-5a2e59ba663c"
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .cuboids["e53bd5e3-980a-4fa7-a0f9-5a2e59ba663c"]
        .name
        == "lidar__cuboid__train"
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .cuboids["e53bd5e3-980a-4fa7-a0f9-5a2e59ba663c"]
        .pos.x
        == 0.0
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .cuboids["e53bd5e3-980a-4fa7-a0f9-5a2e59ba663c"]
        .pos.y
        == 1.0
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .cuboids["e53bd5e3-980a-4fa7-a0f9-5a2e59ba663c"]
        .pos.z
        == 2.0
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .cuboids["e53bd5e3-980a-4fa7-a0f9-5a2e59ba663c"]
        .quat.x
        == 3.0
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .cuboids["e53bd5e3-980a-4fa7-a0f9-5a2e59ba663c"]
        .quat.y
        == 4.0
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .cuboids["e53bd5e3-980a-4fa7-a0f9-5a2e59ba663c"]
        .quat.z
        == 5.0
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .cuboids["e53bd5e3-980a-4fa7-a0f9-5a2e59ba663c"]
        .quat.w
        == 6.0
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .cuboids["e53bd5e3-980a-4fa7-a0f9-5a2e59ba663c"]
        .size.x
        == 7.0
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .cuboids["e53bd5e3-980a-4fa7-a0f9-5a2e59ba663c"]
        .size.y
        == 8.0
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .cuboids["e53bd5e3-980a-4fa7-a0f9-5a2e59ba663c"]
        .size.z
        == 9.0
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .cuboids["e53bd5e3-980a-4fa7-a0f9-5a2e59ba663c"]
        .sensor
        == scene.sensors["lidar"]
    )
    assert scene.frames[1].object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"].cuboids[
        "e53bd5e3-980a-4fa7-a0f9-5a2e59ba663c"
    ].attributes == {"test_bool_attr0": False}


def test_load_frame1_seg3ds(openlabel_v1_short_data, loader):
    scene = loader.load(openlabel_v1_short_data, validate=False)

    assert len(scene.frames[1].object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"].seg3ds) == 2

    assert (
        "550df2c3-0e66-483e-bcc6-f3013b7e581b"
        in scene.frames[1].object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"].seg3ds
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .seg3ds["550df2c3-0e66-483e-bcc6-f3013b7e581b"]
        == scene.frames[1].annotations["550df2c3-0e66-483e-bcc6-f3013b7e581b"]
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .seg3ds["550df2c3-0e66-483e-bcc6-f3013b7e581b"]
        .uid
        == "550df2c3-0e66-483e-bcc6-f3013b7e581b"
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .seg3ds["550df2c3-0e66-483e-bcc6-f3013b7e581b"]
        .name
        == "lidar__vec__train"
    )
    assert scene.frames[1].object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"].seg3ds[
        "550df2c3-0e66-483e-bcc6-f3013b7e581b"
    ].point_ids == list(range(10))
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .seg3ds["550df2c3-0e66-483e-bcc6-f3013b7e581b"]
        .sensor
        == scene.sensors["lidar"]
    )
    assert scene.frames[1].object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"].seg3ds[
        "550df2c3-0e66-483e-bcc6-f3013b7e581b"
    ].attributes == {
        "test_text_attr0": "test_text_attr0_val",
        "test_text_attr1": "test_text_attr1_val",
        "test_num_attr0": 0,
        "test_num_attr1": 1,
        "test_bool_attr0": True,
        "test_vec_attr0": ["0", "1"],
        "test_vec_attr1": [0, 1, 2],
        "test_vec_attr2": ["a", "b", "c"],
    }

    assert (
        "12b21c52-06ea-4269-9805-e7167e7a74ed"
        in scene.frames[1].object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"].seg3ds
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .seg3ds["12b21c52-06ea-4269-9805-e7167e7a74ed"]
        == scene.frames[1].annotations["12b21c52-06ea-4269-9805-e7167e7a74ed"]
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .seg3ds["12b21c52-06ea-4269-9805-e7167e7a74ed"]
        .uid
        == "12b21c52-06ea-4269-9805-e7167e7a74ed"
    )
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .seg3ds["12b21c52-06ea-4269-9805-e7167e7a74ed"]
        .name
        == "lidar__vec__train"
    )
    assert scene.frames[1].object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"].seg3ds[
        "12b21c52-06ea-4269-9805-e7167e7a74ed"
    ].point_ids == list(reversed(range(10)))
    assert (
        scene.frames[1]
        .object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
        .seg3ds["12b21c52-06ea-4269-9805-e7167e7a74ed"]
        .sensor
        == scene.sensors["lidar"]
    )
    assert scene.frames[1].object_data["22dedd49-6dcb-413b-87ef-00ccfb532e98"].seg3ds[
        "12b21c52-06ea-4269-9805-e7167e7a74ed"
    ].attributes == {"test_bool_attr0": False}


def test_load_uri_vcd_incompatible(
    openlabel_v1_short_data, openlabel_v1_vcd_incompatible_data, loader
):
    """Tests, whether an older annotation file, which is not usable for the VCD
    library is converted into a compatible one."""

    scene_ground_truth = loader.load(openlabel_v1_short_data, validate=False)
    scene = loader.load(openlabel_v1_vcd_incompatible_data, validate=False)

    # The UUIDs of the frame data have been generated and therefore do not match the ground truth.
    # They are set equal here.
    for frame_id in scene_ground_truth.frames:
        for frame_data in scene_ground_truth.frames[frame_id].data:
            for other_frame_data in list(scene.frames[frame_id].data):
                if (
                    scene.frames[frame_id].data[other_frame_data].name
                    == scene_ground_truth.frames[frame_id].data[frame_data].name
                ):
                    scene.frames[frame_id].data[frame_data] = scene.frames[frame_id].data[
                        other_frame_data
                    ]
                    del scene.frames[frame_id].data[other_frame_data]
                    scene.frames[frame_id].data[frame_data].uid = frame_data
                    break

    assert scene == scene_ground_truth


# Tests the warnings and errors
def test_no_warnings(openlabel_v1_short_data, loader):
    loader.load(openlabel_v1_short_data, validate=False)
    assert len(loader.warnings) == 0


def test_stream_with_no_coordinate_system(openlabel_v1_short_data, loader):
    del openlabel_v1_short_data["openlabel"]["coordinate_systems"]["ir_middle"]
    del openlabel_v1_short_data["openlabel"]["coordinate_systems"]["base"]["children"][
        openlabel_v1_short_data["openlabel"]["coordinate_systems"]["base"]["children"].index(
            "ir_middle"
        )
    ]
    with pytest.raises(raillabel.exceptions.MissingCoordinateSystemError):
        loader.load(openlabel_v1_short_data)


def test_warnings_sync(openlabel_v1_short_data, loader):
    openlabel_v1_short_data["openlabel"]["frames"]["0"]["frame_properties"]["streams"][
        "non_existing_stream"
    ] = {"stream_properties": {"sync": {"timestamp": "1632321743.100000072"}}}

    loader.load(openlabel_v1_short_data, validate=False)
    assert len(loader.warnings) == 1

    # Tests for keywords in the warning that can help the user identify the source
    assert "frame" in loader.warnings[0]
    assert "0" in loader.warnings[0]
    assert "sync" in loader.warnings[0]
    assert "non_existing_stream" in loader.warnings[0]


def test_warnings_stream_sync_field(openlabel_v1_short_data, loader):
    openlabel_v1_short_data["openlabel"]["frames"]["0"]["frame_properties"]["streams"][
        "rgb_middle"
    ]["stream_properties"]["stream_sync"] = openlabel_v1_short_data["openlabel"]["frames"]["0"][
        "frame_properties"
    ][
        "streams"
    ][
        "rgb_middle"
    ][
        "stream_properties"
    ][
        "sync"
    ]
    del openlabel_v1_short_data["openlabel"]["frames"]["0"]["frame_properties"]["streams"][
        "rgb_middle"
    ]["stream_properties"]["sync"]

    loader.load(openlabel_v1_short_data, validate=False)
    assert len(loader.warnings) == 1

    # Tests for keywords in the warning that can help the user identify the source
    assert "stream_sync" in loader.warnings[0]
    assert "deprecated" in loader.warnings[0].lower()
    assert "save()" in loader.warnings[0]


def test_warnings_ann_object(openlabel_v1_short_data, loader):
    openlabel_v1_short_data["openlabel"]["frames"]["0"]["objects"][
        "affeaffe-0327-46ff-9c28-2506cfd6d934"
    ] = {"object_data": {}}

    loader.load(openlabel_v1_short_data, validate=False)
    assert len(loader.warnings) == 1

    assert not "affeaffe-0327-46ff-9c28-2506cfd6d934" in loader.scene.frames[0].object_data

    # Tests for keywords in the warning that can help the user identify the source
    assert "frame" in loader.warnings[0]
    assert "0" in loader.warnings[0]
    assert "object" in loader.warnings[0]
    assert "affeaffe-0327-46ff-9c28-2506cfd6d934" in loader.warnings[0]


def test_warnings_wrong_annotation_cs(openlabel_v1_short_data, loader):
    openlabel_v1_short_data["openlabel"]["frames"]["0"]["objects"][
        "b40ba3ad-0327-46ff-9c28-2506cfd6d934"
    ]["object_data"]["bbox"][0]["coordinate_system"] = "non_existent_sensor"

    loader.load(openlabel_v1_short_data, validate=False)
    assert len(loader.warnings) == 1

    assert (
        loader.scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .bboxs["78f0ad89-2750-4a30-9d66-44c9da73a714"]
        .sensor
        is None
    )

    # Tests for keywords in the warning that can help the user identify the source
    assert "annotation" in loader.warnings[0]
    assert "78f0ad89-2750-4a30-9d66-44c9da73a714" in loader.warnings[0]
    assert "sensor" in loader.warnings[0]
    assert "non_existent_sensor" in loader.warnings[0]


# Executes the test if the file is called
if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear"])



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
from dataclasses import dataclass, field

from ._annotation import _Annotation
from .bbox import Bbox
from .cuboid import Cuboid
from .object import Object
from .poly2d import Poly2d
from .poly3d import Poly3d
from .seg3d import Seg3d


@dataclass
class ObjectData:
    """Annotations associated with a specific object in a frame.

    Parameters
    ----------
    object: raillabel.format.Object
        A reference to the object this ObjectData belongs to.
    annotations: dict, optional
        Dictionary of all annotations representing this object in this frame.

    Properties (read-only)
    ----------------------
    bboxs: dict of raillabel.format.Bbox
        Dictionary of all bounding boxes representing this object in this frame.
    cuboids: dict of raillabel.format.Cuboid
        Dictionary of all cuboids representing this object in this frame.
    poly2ds: dict of raillabel.format.Poly2d
        Dictionary of all 2d polylines representing this object in this frame.
    poly3ds: dict of raillabel.format.Poly3d
        Dictionary of all 3d polylines representing this object in this frame.
    seg3ds: dict of raillabel.format.Seg3d
        Dictionary of all 3d segmentations representing this object in this frame.
    """

    object: Object
    annotations: t.Dict[str, t.Type[_Annotation]] = field(default_factory=dict)

    @property
    def bboxs(self) -> t.Dict[str, Bbox]:
        """Return dictionary of all bounding boxes."""
        return {k: v for k, v in self.annotations.items() if isinstance(v, Bbox)}

    @property
    def cuboids(self) -> t.Dict[str, Cuboid]:
        """Return dictionary of all cuboids."""
        return {k: v for k, v in self.annotations.items() if isinstance(v, Cuboid)}

    @property
    def poly2ds(self) -> t.Dict[str, Poly2d]:
        """Return dictionary of all 2d poly lines."""
        return {k: v for k, v in self.annotations.items() if isinstance(v, Poly2d)}

    @property
    def poly3ds(self) -> t.Dict[str, Bbox]:
        """Return dictionary of all 3d poly lines."""
        return {k: v for k, v in self.annotations.items() if isinstance(v, Poly3d)}

    @property
    def seg3ds(self) -> t.Dict[str, Seg3d]:
        """Return dictionary of all 3d segmentations."""
        return {k: v for k, v in self.annotations.items() if isinstance(v, Seg3d)}

    def asdict(self) -> dict:
        """Export self as a dict compatible with the OpenLABEL schema.

        Returns
        -------
        dict_repr: dict
            Dict representation of this class instance.

        Raises
        ------
        ValueError
            if an attribute can not be converted to the type required by the OpenLabel schema.
        """

        dict_repr = {"object_data": {}}

        if self.bboxs != {}:
            dict_repr["object_data"]["bbox"] = []

            for bbox in self.bboxs.values():
                dict_repr["object_data"]["bbox"].append(bbox.asdict())

        if self.poly2ds != {}:
            dict_repr["object_data"]["poly2d"] = []

            for poly2d in self.poly2ds.values():
                dict_repr["object_data"]["poly2d"].append(poly2d.asdict())

        if self.poly3ds != {}:
            dict_repr["object_data"]["poly3d"] = []

            for poly3d in self.poly3ds.values():
                dict_repr["object_data"]["poly3d"].append(poly3d.asdict())

        if self.cuboids != {}:
            dict_repr["object_data"]["cuboid"] = []

            for cuboid in self.cuboids.values():
                dict_repr["object_data"]["cuboid"].append(cuboid.asdict())

        if self.seg3ds != {}:
            dict_repr["object_data"]["vec"] = []

            for seg3d in self.seg3ds.values():
                dict_repr["object_data"]["vec"].append(seg3d.asdict())

        return dict_repr


class AnnotationContainer(dict):
    """Advanced version of a dictionary.

    When an element is referenced, it is first searched by its key and
    then by the elements name property. Enables searching by the
    annotation uid and the annotation name.
    """

    @property
    def switched_keys(self):
        """Return the dictionary with the keys switched out to the names."""
        return {v.name: v for v in super().values()}

    def __getitem__(self, __key: str) -> object:
        """Return the item either from the uid or the name."""
        try:
            return super().__getitem__(__key)
        except KeyError as e:
            try:
                return self.switched_keys[__key]
            except KeyError as e1:
                raise e from e1

    def __contains__(self, __o: object) -> bool:
        """Return true if the object is in the dict."""
        if super().__contains__(__o):
            return True
        return self.switched_keys.__contains__(__o)



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
from dataclasses import dataclass

from ._annotation import _Annotation
from .point3d import Point3d


@dataclass
class Poly3d(_Annotation):
    """Sequence of 3D points. Can either be a polygon or polyline.

    Parameters
    ----------
    uid: str
        This a string representing the unique universal identifier for the annotation.
    name: str
        Human readable name describing the annotation.
    points: list of raillabel.format.Point3d
        List of the 3d points that make up the polyline.
    closed: bool
        This parameter states, whether the polyline represents a closed shape (a polygon) or an
        open line.
    attributes: dict, optional
        Attributes of the annotation. Dict keys are the name str of the attribute, values are the
        attribute values. Default is {}.
    sensor: raillabel.format.CoordinateSystem, optional
        A reference to the sensor, this annotation is labeled in. Default is None.
    """

    points: t.List[Point3d] = None
    closed: bool = None

    _REQ_FIELDS = ["points", "closed"]

    @classmethod
    def fromdict(
        self,
        data_dict: dict,
        sensors: dict,
    ) -> t.Tuple["Poly3d", list]:
        """Generate a Bbox object from a dictionary in the OpenLABEL format.

        Parameters
        ----------
        data_dict: dict
            OpenLABEL format dictionary containing the data for the annotation.
        sensors: dict
            Dictionary containing all sensors for the scene.

        Returns
        -------
        annotation: Bbox
            Converted annotation.
        warnings: list of str
            List of non-critical errors, that have occurred during the conversion.
        """

        warnings = []  # list of warnings, that have occurred during the parsing

        # Parses the points
        points = []
        for i in range(0, len(data_dict["val"]), 3):
            points.append(
                Point3d(x=data_dict["val"][i], y=data_dict["val"][i + 1], z=data_dict["val"][i + 2])
            )

        # Creates the annotation with all mandatory properties
        annotation = Poly3d(
            uid=str(data_dict["uid"]),
            name=str(data_dict["name"]),
            closed=data_dict["closed"],
            points=points,
        )

        # Adds the optional properties
        if "coordinate_system" in data_dict and data_dict["coordinate_system"] != "":
            try:
                annotation.sensor = sensors[data_dict["coordinate_system"]]

            except KeyError:
                warnings.append(
                    f"{data_dict['coordinate_system']} does not exist as a coordinate system, "
                    + f"but is referenced for the annotation {data_dict['uid']}."
                )

        # Adds the attributes
        if "attributes" in data_dict:
            annotation.attributes = {
                a["name"]: a["val"] for l in data_dict["attributes"].values() for a in l
            }

        return annotation, warnings

    def asdict(self) -> dict:
        """Export self as a dict compatible with the OpenLABEL schema.

        Returns
        -------
        dict_repr: dict
            Dict representation of this class instance.

        Raises
        ------
        ValueError
            if an attribute can not be converted to the type required by the OpenLabel schema.
        """

        dict_repr = self._annotation_required_fields_asdict()

        dict_repr["closed"] = bool(self.closed)
        dict_repr["val"] = []
        for point in self.points:
            dict_repr["val"].extend(point.asdict())

        dict_repr.update(self._annotation_optional_fields_asdict())

        return dict_repr
